/**
 * forked ntpclient to get date https://github.com/taranais/NTPClient 
 * using the FastLED library!! The adafruit neopixel lib provoked 1-2 hardware
 * resets/hour! 
 */

#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>
#include <NTPClient.h>
#include "Wire.h"
#include "uRTCLib.h"
#include <Adafruit_Sensor.h>
#include <Adafruit_TSL2561_U.h>
//#include <Adafruit_NeoPixel.h>
#include <FastLED.h>
#include "tick_1.h"
#include "xmas_22050.h"

#define wav xmas_22050
#define tick tick
//#define DEBUG // uncomment if DEBUG not wanted

long tick_time_after_boot_s = 30;

ESP8266WiFiMulti Wifi_Multi;      

const byte LED_PIN = 14; //RxD on Wemos D1 mini pro DMA
const byte LED_COUNT = 60;

CRGB Neopix[LED_COUNT];

const long UTC_OFFSET_IN_SECONDS = 3600;
WiFiUDP NTP_UDP; // Define NTP Client to get time
NTPClient NTP_Client(NTP_UDP, "pool.ntp.org", UTC_OFFSET_IN_SECONDS);
uRTCLib RTC(0x68);
Adafruit_TSL2561_Unified TSL = Adafruit_TSL2561_Unified(TSL2561_ADDR_FLOAT, 12345);

byte brightness = 100;
const unsigned long COLOR_SEC  = 0x000001; // blue
const unsigned long COLOR_MIN  = 0x000100; // green
const unsigned long COLOR_HOUR = 0x010000; // red
String formattedDate, dayStamp, timeStamp;
int rtc_day, rtc_month, rtc_year;
int sync_hour = 4;
byte neopix_sec,neopix_min, neopix_hour, neopix_sec_prev,neopix_min_prev, neopix_hour_prev;
unsigned long color_sec, color_min, color_hour, color_sm, color_sh, color_mh, color_smh;
unsigned long micros_now, micros_prev;
unsigned long millis_curr_1, millis_prev_1, millis_curr_2, millis_prev_2;
byte wifi_counter = 0, time_to_wait_for_wifi = 60;

/*******  SETUP **********************************************************************************/  

void setup() {
  #ifdef DEBUG
    Serial.begin(115200);
    Serial.println("Serial OK");
  #endif //DEBUG 
  init_music();   
  init_wifi();      // Try to connect to some given access points. Then wait for a connection  
  FastLED.addLeds<WS2812B, LED_PIN, GRB>(Neopix, LED_COUNT);
  play_song();
  if (wifi_counter < time_to_wait_for_wifi) { // only if wifi found
    NTP_Client.begin();  
    while(!NTP_Client.update()) {
      NTP_Client.forceUpdate();
    }
  }  
  say_hello(500);
  Wire.begin(4, 5); // D3 and D4 on ESP8266
  Wire.setClock(400000L);
  Wire.status();
  if (wifi_counter < time_to_wait_for_wifi) { // only if wifi found
    get_ntp_set_rtc(-1);  
  }  
  RTC.refresh();           
  #ifdef DEBUG
    RTC_print2serial();
  #endif //DEBUG
  init_light_sensor();
}
/******* MAINLOOP **********************************************************************************/  

void loop() {  
  neo_clock();
  play_tick();   
  millis_curr_2 = millis();
  if (millis_curr_2 - millis_prev_2 > 60000) {    
    millis_prev_2 = millis_curr_2;    
    #ifdef DEBUG
      RTC_print2serial();
    #endif //DEBUG      
    set_brightness_from_light();
    if (wifi_counter < time_to_wait_for_wifi) { // only if wifi found
      get_ntp_set_rtc(sync_hour);
    }
  }  
}

/******* INIT FUNCTIONS **********************************************************************************/  

void init_music() {
  analogWriteFreq(22050); // 22kHz PWM
  analogWriteRange(255);  
}

void init_light_sensor() {
  if(!TSL.begin()) {
    #ifdef DEBUG
      Serial.print("Ooops, no TSL2561 detected ... Check your wiring or I2C ADDR!");
    #endif //DEBUG    
    while(1);
  }
  TSL.enableAutoRange(true);                              // Auto-gain
  TSL.setIntegrationTime(TSL2561_INTEGRATIONTIME_13MS);   // fast but low resolution  
}

void init_wifi() { // Try to connect to some given access points. Then wait for a connection
  Wifi_Multi.addAP("yellow", "Fe+30Gu-36Re*38Je/44!");   // add Wi-Fi networks you want to connect to
  Wifi_Multi.addAP("VOO-246590", "GGVWAIOK");
  Wifi_Multi.addAP("ssid_from_AP_3", "your_password_for_AP_3");
  #ifdef DEBUG
    Serial.println("Connecting");
  #endif //DEBUG      
  while (Wifi_Multi.run() != WL_CONNECTED) {  // Wait for the Wi-Fi to connect
    delay(1000);
    wifi_counter++;
    #ifdef DEBUG    
      Serial.print('.');
    #endif //DEBUG
    if (wifi_counter == time_to_wait_for_wifi) {
      break;   
    }
  }
  #ifdef DEBUG
    Serial.println("\r\n");
    Serial.print("Connected to ");
    Serial.println(WiFi.SSID());             // Tell us what network we're connected to
    Serial.print("IP address:\t");
    Serial.print(WiFi.localIP());            // Send the IP address of the ESP8266 to the computer
    Serial.println("\r\n");
  #endif //DEBUG    
}

/******* HELPER FUNCTIONS ******************************************************/
void setall(byte R, byte G, byte B) {  
  for(int i=0;i<LED_COUNT;i++) {    
    Neopix[i].r = R;
    Neopix[i].g = G;
    Neopix[i].b = B;
  }
  FastLED.show();
}

void say_hello(int delay_val) {
  setall(20,0,0);
  delay(delay_val);
  setall(0,20,0);
  delay(delay_val);
  setall(0,0,20);
  delay(delay_val);
  setall(20,20,0);
  delay(delay_val);
  setall(0,20,20);
  delay(delay_val);
  setall(20,0,20);
  delay(delay_val);
  setall(20,20,20);
  delay(delay_val);
  setall(0,0,0);
}  

void neo_clock() {
  RTC.refresh();  
  color_sec  = COLOR_SEC * brightness;
  color_min  = COLOR_MIN * brightness;
  color_hour = COLOR_HOUR * brightness;
  color_sm   = color_sec + color_min;
  color_sh   = color_sec + color_hour;
  color_mh   = color_min + color_hour;
  color_smh  = color_sec + color_min + color_hour; 
  //Clear previous LEDs
  Neopix[neopix_sec_prev] = CRGB::Black;
  Neopix[neopix_min_prev] = CRGB::Black;
  Neopix[neopix_hour_prev] = CRGB::Black;
  neopix_sec = RTC.second();
  if (neopix_sec == 0) { 
    neopix_sec = 59;    
  }
  else {
    neopix_sec -= 1;    
  }  
  neopix_min = RTC.minute();
  if (neopix_min == 0) { 
    neopix_min = 59;    
  }
  else {
    neopix_min -= 1;    
  }
  neopix_hour = RTC.hour();
  if (neopix_hour >= 12) {
    neopix_hour -= 12;
  }
  neopix_hour *= 5;
  neopix_hour += (neopix_min/12);
  neopix_sec_prev = neopix_sec;
  neopix_min_prev = neopix_min;
  neopix_hour_prev = neopix_hour;
  if ((neopix_sec == neopix_min) && (neopix_sec == neopix_hour)) {
    Neopix[neopix_sec] = color_smh;  
    Neopix[neopix_min] = color_smh;  
    Neopix[neopix_hour] = color_smh;  
  }
  else if (neopix_sec == neopix_min) {
    Neopix[neopix_sec] = color_sm;  
    Neopix[neopix_min] = color_sm;  
    Neopix[neopix_hour] = color_hour;  
  }
  else if (neopix_sec == neopix_hour) {
    Neopix[neopix_sec] = color_sh;  
    Neopix[neopix_min] = color_min;  
    Neopix[neopix_hour] = color_sh;  
  }
  else if (neopix_min == neopix_hour) {
    Neopix[neopix_sec] = color_sec;  
    Neopix[neopix_min] = color_mh;  
    Neopix[neopix_hour] = color_mh;  
  }
  else {  
    Neopix[neopix_sec] = color_sec;  
    Neopix[neopix_min] = color_min;  
    Neopix[neopix_hour] = color_hour;  
  }
  FastLED.show(); // Send the updated pixel colors to the hardware.  
  delay(20);
}

void RTC_print2serial() {
  Serial.print("RTC DateTime: ");
  Serial.print(RTC.year());
  Serial.print('/');
  Serial.print(RTC.month());
  Serial.print('/');
  Serial.print(RTC.day());
  Serial.print(' ');
  Serial.print(RTC.hour());
  Serial.print(':');
  Serial.print(RTC.minute());
  Serial.print(':');
  Serial.print(RTC.second());
  Serial.print(" Day: ");
  Serial.print(RTC.dayOfWeek());
  Serial.print(" - Temp: ");
  Serial.println(RTC.temp()  / 100);  
}

void play_song() {
  int count = 0;
  micros_prev = micros();
  while (count != sizeof(wav)) {
    micros_now = micros();
    if ((micros_now - micros_prev) > 45) {
      micros_prev = micros_now;
      byte a_value = pgm_read_byte(wav+count); 
      analogWrite(D7,a_value);   
      count++;
      yield(); // to avoid watchdog reset
    }
  }
}

void play_tick() {
  if (tick_time_after_boot_s > 0) {
    millis_curr_1 = millis();
    if (millis_curr_1 - millis_prev_1 > 1000) {
      millis_prev_1 = millis_curr_1;
      tick_time_after_boot_s--;
      for (long j= 0; j<sizeof(tick); j++) {  // play tick
        byte a_value = pgm_read_byte(tick+j); 
        analogWrite(D7,a_value);       
      }
      yield(); 
     }
  }
}  

void set_brightness_from_light() {
  sensors_event_t event;
  TSL.getEvent(&event);
  if (event.light) {    
    if (event.light > 255) {    
      event.light = 255;
    }
    brightness = map(event.light,1,255,20,150);
    #ifdef DEBUG
      Serial.print(event.light);
      Serial.println(" lux");    
      Serial.print("Brightness: ");
      Serial.println(brightness);
    #endif //DEBUG    
  }
  else {
    #ifdef DEBUG
      Serial.println("Sensor overload");
    #endif //DEBUG    
    brightness = 20;
  }
}

void get_ntp_set_rtc(int sync_hour) { // if -1 ignore sync_hour
  while(!NTP_Client.update()) {
    NTP_Client.forceUpdate();
  }
  formattedDate = NTP_Client.getFormattedDate();        
  int splitT = formattedDate.indexOf("T"); // Extract date
  dayStamp = formattedDate.substring(0, splitT);
  timeStamp = formattedDate.substring(splitT+1, formattedDate.length()-1);
  rtc_day = dayStamp.substring(8).toInt();
  rtc_month = dayStamp.substring(5,7).toInt();
  rtc_year = dayStamp.substring(2,4).toInt();
  #ifdef DEBUG
    Serial.print("DATE: ");
    Serial.println(dayStamp);  
    Serial.print("HOUR: ");
    Serial.println(timeStamp);        
  #endif //DEBUG    
  if (NTP_Client.update() && (sync_hour == -1)) {  
    RTC.set(NTP_Client.getSeconds(), NTP_Client.getMinutes(), NTP_Client.getHours(),
    NTP_Client.getDay(), rtc_day, rtc_month, rtc_year);
    #ifdef DEBUG
      Serial.println("RTC update successfull!");
    #endif //DEBUG          
  }
  else if (NTP_Client.update() && (RTC.hour()==sync_hour) && (RTC.minute()==0 || RTC.minute()==1)) {
    RTC.set(NTP_Client.getSeconds(), NTP_Client.getMinutes(), NTP_Client.getHours(),
    NTP_Client.getDay(), rtc_day, rtc_month, rtc_year);
    #ifdef DEBUG
      Serial.println("RTC update successfull!");
    #endif //DEBUG          
  }
  delay(20);
}          
  
